CONFIG_ALL = {
    "backbone": "spacy",
    "language": "en",
    "model": "en_core_web_sm",
    "text_column": "text",
    "max_length": 1000000,
    "remove_constant_cols": True,
    "features": {
        "surface": [
            "raw_sequence_length",
            "n_tokens",
            "n_sentences",
            "n_tokens_per_sentence",
            "n_characters",
            "avg_word_length",
            "n_types",
            "n_long_words",
            "n_lemmas"
        ],
        "morphology": ["n_per_morph_feature"],
        "dependency": [
            "tree_width",
            "tree_depth",
            "tree_branching",
            "n_noun_chunks",
            "n_per_dependency_type"
        ],
        "emotion": [
            "sentiment_score",
            "n_negative_sentiment",
            "n_positive_sentiment",
            "avg_valence",
            "n_low_valence",
            "n_high_valence",
            "avg_arousal",
            "n_low_arousal",
            "n_high_arousal",
            "avg_dominance",
            "n_low_dominance",
            "n_high_dominance",
            "avg_emotion_intensity",
            "n_low_intensity",
            "n_high_intensity",
        ],
        "information": [
            "compressibility",
            "entropy"
        ],
        "lexical_richness": [
            "lemma_token_ratio",
            "ttr",
            "rttr",
            "cttr",
            "herdan_c",
            "summer_index",
            "dougast_u",
            "maas_index",
            "lexical_density",
            "n_hapax_legomena",
            "n_global_token_hapax_legomena",
            "n_global_lemma_hapax_legomena",
            "n_hapax_dislegomena",
            "n_global_token_hapax_dislegomena",
            "n_global_lemma_hapax_dislegomena",
            "sichel_s",
            "global_sichel_s"
            "giroud_index",
            "mtld",
            "hdd",
            "mattr",
            "msttr",
            "yule_k",
            "simpsons_d",
            "herdan_v",
        ],
        "pos": [
            "n_lexical_tokens",
            "pos_variability",
            "n_per_pos",
        ],
        "psycholinguistic": [
            "avg_concreteness",
            "n_low_concreteness",
            "n_high_concreteness",
            "avg_sd_concreteness",
            "n_controversial_concreteness",
            "avg_aoa",
            "n_low_aoa",
            "n_high_aoa",
            "avg_sd_aoa",
            "n_controversial_aoa",
            "avg_prevalence",
            "n_low_prevalence",
            "n_high_prevalence",
            "avg_socialness",
            "n_low_socialness",
            "n_high_socialness",
            "avg_sd_socialness",
            "n_controversial_socialness",
            "avg_sensorimotor",
            "n_low_sensorimotor",
            "n_high_sensorimotor",
            "avg_sd_sensorimotor",
            "n_controversial_sensorimotor",
            "avg_iconicity",
            "n_low_iconicity",
            "n_high_iconicity",
            "avg_sd_iconicity",
            "n_controversial_iconicity",
        ],
        "readability": [
            "n_syllables",
            "n_monosyllables",
            "n_polysyllables",
            "flesch_reading_ease",
            "flesch_kincaid_grade",
            "gunning_fog",
            "ari",
            "smog",
            "cli",
            "lix",
            "rix",
        ],
        "semantic": [
            "n_lemmas",
            "n_hedges",
            "avg_num_synsets",
            "avg_num_synsets_per_pos",
            "n_low_synsets",
            "n_high_synsets",
            "n_high_synsets_per_pos",
            "n_low_synsets_per_pos",
        ],
        "entities": [
            "n_entities",
            "n_entities_token_ratio",
            "n_entities_sentence_ratio",
            "n_per_entity_type",
        ],
    }
}